/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 * changecase.c
 * This file is part of gedit
 *
 * Copyright (C) 2002 Paolo Borelli 
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, 
 * Boston, MA 02111-1307, USA. 
 */
 

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <libgnome/gnome-i18n.h>
#include <libgnome/gnome-help.h>

#include <gedit-plugin.h>
#include <gedit-debug.h>
#include <gedit-menus.h>
#include <gedit-utils.h>


#define SUBMENU_LABEL	N_("C_hange Case")
#define SUBMENU_NAME	"ChangeCase"	
#define SUBMENU_PATH	"/menu/Edit/EditOps_6/"

#define MENU_ITEMS_PATH	"/menu/Edit/EditOps_6/ChangeCase/"

#define UPPER_ITEM_LABEL	N_("All _Upper Case")
#define LOWER_ITEM_LABEL	N_("All _Lower Case")
#define INVERT_ITEM_LABEL	N_("_Invert Case")
#define TITLE_ITEM_LABEL	N_("_Title Case")

#define UPPER_ITEM_TIP		N_("Change selected text to upper case")
#define LOWER_ITEM_TIP		N_("Change selected text to lower case")
#define INVERT_ITEM_TIP		N_("Inverts the case of selected text")
#define TITLE_ITEM_TIP		N_("Capitalize the first letter of each word selected")


G_MODULE_EXPORT GeditPluginState update_ui (GeditPlugin *plugin, BonoboWindow *window);
G_MODULE_EXPORT GeditPluginState init (GeditPlugin *pd);
G_MODULE_EXPORT GeditPluginState destroy (GeditPlugin *pd);
G_MODULE_EXPORT GeditPluginState activate (GeditPlugin *pd);
G_MODULE_EXPORT GeditPluginState deactivate (GeditPlugin *pd);


typedef enum {
	TO_UPPER_CASE,
	TO_LOWER_CASE,
	INVERT_CASE,
	TITLE_CASE,
} ChangeCaseChoice;


static gchar*
do_upper_case (const gchar *text)
{
	static gchar *ret_text;

	ret_text = g_utf8_strup (text, -1);
	return ret_text;	
}


static gchar*
do_lower_case (const gchar *text)
{
	static gchar *ret_text;

	ret_text = g_utf8_strdown (text, -1);
	return ret_text;	
}


static gchar*
do_invert_case (const gchar *text)
{
	const gchar *p = text;
	static gchar *ret_text;

	GString *s = NULL;
	s = g_string_new ("\0");

	while (*p) {
		gunichar c = g_utf8_get_char (p);
		gunichar newc;

		if (g_unichar_islower (c))
			newc = g_unichar_toupper (c);
		else
			newc = g_unichar_tolower (c);

		s = g_string_append_unichar (s, newc);
		
		p = g_utf8_next_char (p);
	}
	
	ret_text = s->str;

	g_string_free (s, FALSE);

	return ret_text;	
}


/*
 * FIXME: this is implemented in a quite stupid way, but I'm not sure how I
 *        should handle it: what about the case when the user selection starts
 *        in the middle of a word?
 *        At the moment it blindly converts the first char of the selection and
 *        every char following a space or a \n; this covers the case of consecutive
 *        spaces since the title case of a space is still a space.
 */
static gchar*
do_title_case (const gchar *text)
{
	const gchar *p = text;
	static gchar *ret_text;
	gunichar c, newc, nextc;

	GString *s = NULL;
	s = g_string_new ("\0");

	/* unconditionally change the first one to title case */
	c = g_utf8_get_char (p);
	newc = g_unichar_totitle (c);
	s = g_string_append_unichar (s, newc);

	while (*p) {
		c = g_utf8_get_char (p);
		p = g_utf8_next_char (p);

		if (!*p)
			break;

		nextc = g_utf8_get_char (p);

		if ((c == ' ') || (c == '\n')) {
			newc = g_unichar_totitle (nextc);
			s = g_string_append_unichar (s, newc);
		} else
			s = g_string_append_unichar (s, nextc);

	}
	
	ret_text = s->str;

	g_string_free (s, FALSE);

	return ret_text;	
}


static void
change_case (ChangeCaseChoice choice)
{
	GeditDocument *doc;
	GeditView *view;
	gint start;
	gint end;
	gchar *buffer;
	gchar *buffer_utf8;
	gchar *changed_text;

	gedit_debug (DEBUG_PLUGINS, "");

	view = gedit_get_active_view ();
	g_return_if_fail (view != NULL);

	doc = gedit_view_get_document (view);
	g_return_if_fail (doc != NULL);

	if (!gedit_document_get_selection (doc, &start, &end))
		/* Should warn that no text is selected? */
		return;
	
	buffer = gedit_document_get_chars (doc, start, end);
	
	if (g_utf8_validate (buffer, -1, NULL))
		buffer_utf8 = buffer;
	else {
		buffer_utf8 = g_locale_to_utf8 (buffer, -1, NULL, NULL, NULL);
		g_free (buffer);
	}

	switch (choice) {
		case TO_UPPER_CASE:
			changed_text = do_upper_case (buffer_utf8);
			break;

		case TO_LOWER_CASE:
			changed_text = do_lower_case (buffer_utf8);
			break;

		case INVERT_CASE:
			changed_text = do_invert_case (buffer_utf8);
			break;

		case TITLE_CASE:
			changed_text = do_title_case (buffer_utf8);
			break;

		default:
			/* Should never get here */
			g_return_if_reached();

	}

	/* not sure if this should be really undoable */
	gedit_document_begin_not_undoable_action (doc);

	gedit_document_delete_text (doc, start, end);
	gedit_document_set_cursor (doc, start);
	gedit_document_insert_text_at_cursor (doc, changed_text, -1);

	gedit_document_end_not_undoable_action (doc);

	g_free (buffer_utf8);
	g_free (changed_text);

}


static void
upper_case_cb (BonoboUIComponent *uic, gpointer user_data, const gchar* verbname)
{
	change_case (TO_UPPER_CASE);
}


static void
lower_case_cb (BonoboUIComponent *uic, gpointer user_data, const gchar* verbname)
{
	change_case (TO_LOWER_CASE);
}


static void
invert_case_cb (BonoboUIComponent *uic, gpointer user_data, const gchar* verbname)
{
	change_case (INVERT_CASE);
}


static void
title_case_cb (BonoboUIComponent *uic, gpointer user_data, const gchar* verbname)
{
	change_case (TITLE_CASE);
}


gchar* changecase_sensible_verbs [] = {
	"/commands/AllUpperCase",
	"/commands/AllLowerCase",
	"/commands/InvertCase",
	"/commands/TitleCase",

	NULL
};


G_MODULE_EXPORT GeditPluginState
update_ui (GeditPlugin *plugin, BonoboWindow *window)
{
	BonoboUIComponent *uic;
	GeditDocument *doc;
	
	gedit_debug (DEBUG_PLUGINS, "");
	g_return_val_if_fail (window != NULL, PLUGIN_ERROR);

	uic = gedit_get_ui_component_from_window (window);
	doc = gedit_get_active_document ();

	if ((doc == NULL) || (gedit_document_is_readonly (doc)))
		gedit_menus_set_verb_list_sensitive (uic, changecase_sensible_verbs, FALSE);
	else
		gedit_menus_set_verb_list_sensitive (uic, changecase_sensible_verbs, TRUE);

	return PLUGIN_OK;
}


G_MODULE_EXPORT GeditPluginState
activate (GeditPlugin *pd)
{
	GList *top_windows;
	gedit_debug (DEBUG_PLUGINS, "");

	top_windows = gedit_get_top_windows ();
	g_return_val_if_fail (top_windows != NULL, PLUGIN_ERROR);

	while (top_windows)
	{

		gedit_menus_add_submenu (BONOBO_WINDOW (top_windows->data),
		                         SUBMENU_PATH, SUBMENU_NAME,
		                         SUBMENU_LABEL);

		gedit_menus_add_menu_item (BONOBO_WINDOW (top_windows->data),
		                           MENU_ITEMS_PATH, "AllUpperCase",
		                           UPPER_ITEM_LABEL, UPPER_ITEM_TIP,
		                           NULL, upper_case_cb);

		gedit_menus_add_menu_item (BONOBO_WINDOW (top_windows->data),
		                           MENU_ITEMS_PATH, "AllLowerCase",
		                           LOWER_ITEM_LABEL, LOWER_ITEM_TIP,
		                           NULL, lower_case_cb);

		gedit_menus_add_menu_item (BONOBO_WINDOW (top_windows->data),
		                           MENU_ITEMS_PATH, "InvertCase",
		                           INVERT_ITEM_LABEL, INVERT_ITEM_TIP,
		                           NULL, invert_case_cb);

		gedit_menus_add_menu_item (BONOBO_WINDOW (top_windows->data),
		                           MENU_ITEMS_PATH, "TitleCase",
		                           TITLE_ITEM_LABEL, TITLE_ITEM_TIP,
		                           NULL, title_case_cb);

                pd->update_ui (pd, BONOBO_WINDOW (top_windows->data));

                top_windows = g_list_next (top_windows);
        }

	return PLUGIN_OK;
}


G_MODULE_EXPORT GeditPluginState
deactivate (GeditPlugin *pd)
{
	gedit_menus_remove_submenu_all (SUBMENU_PATH, SUBMENU_NAME);

	return PLUGIN_OK;
}


G_MODULE_EXPORT GeditPluginState
init (GeditPlugin *pd)
{
	/* initialize */
	gedit_debug (DEBUG_PLUGINS, "");
		
	return PLUGIN_OK;
}


G_MODULE_EXPORT GeditPluginState
destroy (GeditPlugin *plugin)
{
	gedit_debug (DEBUG_PLUGINS, "");

	return PLUGIN_OK;
}

